@@ -27,6 +27,7 @@
 #		zskphase	"2"
 #		kskphase	"0"
 #		maxttl		"86400"
+#		curerrors	"0"
 #		display		"0"
 #		phasestart	"Wed Mar 09 21:49:22 2005"
 #
@@ -39,6 +40,8 @@
 #		maxttl		"100000"
 #		display		"1"
 #		loglevel	"info"
+#		curerrors	"0"
+#		maxerrors	"5"
 #		phasestart	"Sun Jan 01 16:00:00 2005"
 #
 #
@@ -62,7 +65,7 @@ use Fcntl qw(:DEFAULT :flock);
 use Net::DNS::SEC::Tools::conf;
 use Net::DNS::SEC::Tools::rollmgr;
 
-our $VERSION = "1.5";
+our $VERSION = "1.8";
 
 our @ISA = qw(Exporter);
 
@@ -116,6 +119,8 @@ my @ROLLFIELDS = (
 			'kskphase',
 			'zskphase',
 			'zsargs',
+			'curerrors',
+			'maxerrors',
 			'ksk_rollsecs',
 			'ksk_rolldate',
 			'zsk_rollsecs',
@@ -187,7 +192,7 @@ sub rollrec_unlock
 # print "rollrec_unlock:  down in\n";
 
 	#
-	# Lock the lock file.
+	# Unlock the lock file.
 	#
 	return(flock(RRLOCK,LOCK_UN));
 }
@@ -1125,6 +1130,12 @@ sub rollrec_close
 # Routine:	rollrec_write()
 #
 # Purpose:	Save the roll record file and leave the file handle open.
+#		We'll get an exclusive lock on the rollrec file in order
+#		to (try to) ensure we're the only ones writing the file.
+#
+#		We'll make a (hopefully atomic) copy of the in-core rollrec
+#		lines prior to trying to write.  This is an attempt to
+#		keep the data from being mucked with while we're using it.
 #
 sub rollrec_write
 {
@@ -1132,6 +1143,9 @@ sub rollrec_write
 	my $rrc = "";		# Concatenated rollrec file contents.
 	my $ofh;		# Old file handle.
 
+	my @rrlines = @rollreclines;	# Copy of The Rollrec.
+	my $rrlen;			# Number of lines in The Rollrec.
+
 # print STDERR "rollrec_write:  down in\n";
 
 	my @currentcmds = rollmgr_getallqueuedcmds();
@@ -1139,27 +1153,45 @@ sub rollrec_write
 	#
 	# If the file hasn't changed, we'll skip writing.
 	#
-	return if(!$modified && $#currentcmds == -1);
+	return if(!$modified && ($#currentcmds == -1));
+
+	#
+	# Make sure we've got the correct count of rollrec lines.
+	#
+	$rrlen = @rrlines;
 
 	#
 	# Loop through the array of rollrec lines and concatenate them all.
 	#
-	for(my $ind=0;$ind<$rollreclen;$ind++)
+	for(my $ind = 0; $ind < $rrlen; $ind++)
 	{
-		$rrc .= $rollreclines[$ind];
+		$rrc .= $rrlines[$ind];
 	}
 
 	#
-	# remember any unprocessed queue commands
+	# Remember any unprocessed queue commands.
 	#
-	if ($writecmds) {
-	    foreach my $cmdandvalue (rollmgr_getallqueuedcmds()) {
-		my ($cmd, $value) = @$cmdandvalue;
-		$cmd =~ s/^rollcmd_//;
-		$rrc .= "cmd \"$cmd $value\"\n";
-	    }
+	if($writecmds)
+	{
+		foreach my $cmdandvalue (rollmgr_getallqueuedcmds())
+		{
+			my ($cmd, $value) = @$cmdandvalue;
+			$cmd =~ s/^rollcmd_//;
+			$rrc .= "cmd \"$cmd $value\"\n";
+		}
 	}
 
+	#
+	# Lock the rollrec file.
+	#
+	flock(ROLLREC,LOCK_EX);
+
+	#
+	# Force immediate writes of ROLLREC.
+	#
+	$ofh = select ROLLREC;
+	$| = 1;
+
 	#
 	# Zap the rollrec file and write out the new one.
 	#
@@ -1168,12 +1200,14 @@ sub rollrec_write
 	print ROLLREC $rrc;
 
 	#
-	# Flush the ROLLREC buffer.
+	# Reset ROLLREC buffering to original state.
 	#
-	$ofh = select ROLLREC;
-	$| = 1;
-
 	select $ofh;
+
+	#
+	# Unlock the rollrec file.
+	#
+	return(flock(ROLLREC,LOCK_UN));
 }
 
 #--------------------------------------------------------------------------
@@ -1300,6 +1334,8 @@ of keyword/value entries.  The following is an example of a I<rollrec>:
 	loglevel		"info"
 	rollrec_rollsecs	"1115923362"
 	rollrec_rolldate	"Tue Mar 09 19:12:54 2005"
+	curerrors		0
+	maxerrors		5
         # optional records:
         istrustanchor           "yes"
         holddowntime            "8W"
@@ -1533,6 +1569,8 @@ measure, an internal modification flag is checked prior to writing the file.
 If the program has not modified the contents of the I<rollrec> file, it is not
 rewritten.
 
+I<rollrec_write()> gets an exclusive lock on the I<rollrec> file while writing.
+
 =back
 
 =head1 ROLLREC INTERNAL INTERFACES
